"""
         Yasuharu Okamoto, 2021.12, 2022.01
"""
import csv
from copy import deepcopy
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sb
from multiprocessing.pool import Pool
import mymodule as mm     #  a personal module, file name: mymodule.py

if __name__ == '__main__':
    fin_nm = input('Input file (*.csv) = ')
    with open(fin_nm, 'r') as f:
        Data_in = [v for v in csv.reader(f)]
        
    Data_in = np.array(Data_in)
    Rs = [int(v) for v in Data_in.T[1][1:]]
    Ags = [float(v) for v in Data_in.T[2][1:]]
 
    print(len(Rs), len(Ags))
    Rs = np.array(Rs)
    Ags = np.array(Ags)
    Ags_raw = deepcopy(Ags)
    Ags_mean = Ags.mean()
    Ags_std = Ags.std()
    Ags = (Ags - Ags_mean) / Ags_std
    #   Construction of a dictionary for d's and probabilities of correct response
    print("Preparing the dictionary of d's and probabilties")
    TblPDprime = mm.calcPDprimes()     

    #
    #      Adjusting the proposal distributions
    #
    i_step = -1
    sgm_a = 0.5
    sgm_b = 0.5
    init_a = 0.0
    init_b = 0.0
    print('Adjusting the proposals')
    while True:
        print('step-{0:} started.'.format(-i_step))
        my_mcmc = mm.MCMC(deepcopy(Rs), deepcopy(Ags), deepcopy(TblPDprime), init_a, init_b, sgm_a, sgm_b, 2000, i_step)
        n_accpt_a, n_accpt_b, L_chain, chain_a, chain_b = my_mcmc.do_mcmc()
        acpt_r_a = n_accpt_a / L_chain
        acpt_r_b = n_accpt_b / L_chain
                                 
        ck_a = 0
        v, ck_a = mm.scale_sgm(acpt_r_a, ck_a)  #   Check the acceptance proportions
        init_a = chain_a.mean()
        if ck_a > 0:
            sgm_a *= v    #   Adjust the sigma of the normal distribution for proposal
        
        ck_b = 0
        v, ck_b = mm.scale_sgm(acpt_r_b, ck_b)  
        init_b = chain_b.mean()
        if ck_b > 0:
            sgm_b *= v
            
        if ck_a + ck_b == 0:   #  The sizes of the sigmas OK?
            break              #  Adjustment ends,
        i_step -= 1
    
    n_samples = 5000      #    Sampling size for each chain in MCMC sampling
            
    #
    #         Parameter values for multiproccing, each of which runs one chain in MCMC
    #
    params_t = {'Rs': deepcopy(Rs), 'Ags': deepcopy(Ags), 'TblPDprime': deepcopy(TblPDprime),
                'init_a': init_a, 'init_b': init_b, 'sgm_a': sgm_a, 'sgm_b': sgm_b,
                   'n_samples': n_samples, 'chain_id': None}
    #
    #    List of parameter values for multiprocessing
    #
    params = []
    for i in range(4):
        p_t = deepcopy(params_t)
        p_t['chain_id'] = i
        params.append(p_t)

    print('MCMC sampling by multiprocessing started.')  
    with Pool() as pool:
        Results = pool.map(mm.MCMCs, params)    #  Run the multiprocessing
    
    print('MCMC sampling ended.')
    
    fout_nm = 'regrslts.txt'
    fout = open(fout_nm, 'w')   #   Output text file

    #
    #       Posterior distributon and trace of the parameter a
    #
    plt.figure(figsize = (12,5))
    plt.subplot(1,2,1)
    param_a = [[]]*4
    for i in range(4):
        param_a[i] = Results[i][3]
    for i in range(4):
        sb.kdeplot(param_a[i])     #  KDE posteriors for the parameter a 
    plt.xlabel('a', fontsize = 16)
    plt.yticks([])
    plt.title('Posterior distributions of a', fontsize = 14)
            
    Rhat_a = mm.calcRhat(param_a)
    plt.subplot(1,2,2)
    for i in range(4):
        plt.plot(range(len(param_a[0])), param_a[i])     #   Traces of sampling of the parameter a
    plt.xlabel('trial', fontsize = 14)
    plt.title(f'Traces of a,  Rhat = {Rhat_a:.2f}', fontsize = 14)
    plt.savefig('FigDistTrace_a.png')
    plt.show()

    plt.figure(figsize=(12,5))
    plt.subplot(1,2,1)
    param_b = [[]] * 4
    for i in range(4):
        param_b[i] = Results[i][4]
    for i in range(4):
        sb.kdeplot(param_b[i])
    plt.xlabel('b', fontsize = 16)
    plt.yticks([])
    plt.title('Posterior distributions of b', fontsize = 14)
            
    Rhat_b = mm.calcRhat(param_b)
    plt.subplot(1,2,2)
    for i in range(4):
        plt.plot(range(len(param_b[i])), param_b[i])
    plt.xlabel('trial', fontsize = 14)
    plt.title(f'Traces of b,  Rhat = {Rhat_b:.2f}', fontsize = 14)
    plt.savefig('FigDistTrace_b.png')
    plt.show()

    a_whole = []
    b_whole = []
    for i in range(4):
        a_whole += [v for v in param_a[i]]
        b_whole += [v for v in param_b[i]]
        
    a_mode = mm.EstMode(a_whole)
    b_mode = mm.EstMode(b_whole)
    print(a_mode, b_mode)

    #       Posterior distribution of a
    sb.kdeplot(a_whole)
    plt.title(f'Posterior distribution of a,  mode = {a_mode:.2f}', fontsize = 16)
    plt.xlabel('a', fontsize = 16)
    plt.yticks([])
    plt.savefig('Fig_a.png')
    plt.show()

    #       Posterior distribution of b
    sb.kdeplot(b_whole)
    plt.title(f'Posterior distribution of b,  mode = {b_mode:.2f}', fontsize = 16)
    plt.xlabel('b', fontsize = 16)
    plt.yticks([])
    plt.savefig('Fig_b.png')
    plt.show()

    ags_min = np.min(Ags)
    ags_max = np.max(Ags)

    def catAgs(v):
        """
                Categorization into 4 categories
        """
        tv = int((v - ags_min) * 4.0 / (ags_max - ags_min))
        if tv > 3:
            tv = 3
        return tv

    x_coord_z = np.linspace(ags_min, ags_max, 101)   #  Grid for x-axis
    y_coord_z = np.empty(len(x_coord_z))
    for i, v in enumerate(x_coord_z):
        d = np.exp(a_mode + b_mode * v)              #  d' predicted by the regression model 
        y_coord_z[i] = mm.f_Pc(d)                    #  Predicted probability
    plt.plot(x_coord_z, y_coord_z, label = 'Model')
    ags_count = np.zeros(4)
    r_counts = np.zeros(4)       #   Number of responses in each of 4 categories 
    rc_counts = np.zeros(4)      #   Number of crrect responses in each of 4 categories
    for rv, av in zip(Rs, Ags):
        ic = catAgs(av)
        r_counts[ic] += 1
        if rv == 1:
            rc_counts[ic] += 1
    ave_rc = rc_counts / r_counts         #  proportions of crrect responses in each category
    ave_ags = np.empty(4)                 #  means of the independent variable in each category
    step_ags = (ags_max - ags_min) / 8.0  #  half of the category width
    for i in range(4):
        ave_ags[i] = ags_min + step_ags + i * step_ags * 2   #  mid points of the categories
    plt.plot(ave_ags, ave_rc, 'o', label = 'data')
    #
    #     Conversion of standardized scores back to raw scores 
    #
    x_ags = [(ags_min + i*step_ags*2) for i in range(5)]   #   positions of the raw score labels
    x_ags = np.array(x_ags)
    x_ags_raw = x_ags * Ags_std + Ags_mean
    x_ags_lbl = [f"{int(v)}" for v in x_ags_raw]                #   labels for raw scores
    plt.xticks(x_ags, x_ags_lbl)
    plt.xlabel('Independent variable', fontsize = 14)
    plt.ylabel('Probability', fontsize = 14)
    plt.title('Regression Curve of Probability', fontsize = 16)
    plt.legend()
    plt.savefig('Fig_regression_p.png')
    plt.show()
    
    fout.write('\nRegression model of Pc on the standardized independent variable Z:\n')
    fout.write('P(R = 1) = Pc(exp({0:.2f} + {1:.2f} * Z))\n'.format(a_mode, b_mode))
    
    a_raw = a_mode - b_mode * Ags_mean / Ags_std
    b_raw = b_mode / Ags_std
    fout.write('\nRegression model of Pc on the raw independent variable X:\n')
    fout.write('P(R = 1) = Pc(exp({0:.2f} + {1:.2f} * X))\n'.format(a_raw, b_raw))
    
    x_coord_z = np.linspace(ags_min, ags_max, 101)   #  Grid for x-axis
    y_coord_d = np.empty(len(x_coord_z))
    for i, v in enumerate(x_coord_z):
        y_coord_d[i] = np.exp(a_mode + b_mode * v)              #  d' predicted by the regression model 
    plt.plot(x_coord_z, y_coord_d)
    plt.xticks(x_ags, x_ags_lbl)
    plt.xlabel('Independent variable', fontsize = 14)
    plt.ylabel("d'", fontsize = 16)
    plt.title("Regression curve of d'", fontsize = 16)
    plt.savefig('Fig_regression_d.png')
    plt.show()

    fout.write("\nRegression model of d' on the standardized independent variable Z:\n")
    fout.write("d' = exp({0:.2f} + {1:.2f} * Z)\n".format(a_mode, b_mode))

    fout.write("\nRegression model of d' on the raw independent variable X:\n")
    fout.write("d' = exp({0:.2f} + {1:.2f} * X)\n".format(a_raw, b_raw))
    
    fout.close()
    print('\n', fout_nm, 'was saved.\n')
